from collections.abc import Iterable
from uuid import UUID, uuid4

from sqlalchemy import (
    BIGINT,
    Column,
    Engine,
    MetaData,
    SmallInteger,
    String,
    Table,
    UniqueConstraint,
    Uuid,
    select,
    text,
    tuple_
)
from sqlalchemy.orm import DeclarativeBase, Session, sessionmaker
from sqlalchemy.dialects.postgresql import insert as pg_insert

from dedup_pg.backend.backend import Backend


class SQLAlchemyBackend(Backend):
    def __init__(
        self,
        *,
        engine: Engine | None = None,
        session_factory: sessionmaker[Session] | None = None,
        base_or_metadata: type[DeclarativeBase] | MetaData,
        table_name: str,
    ) -> None:
        """
        The SQLAlchemy backend for the deduplication indexing layer.

        Note that this expects a PostgreSQL engine or session_factory to be provided.

        Args:
            base (type[DeclarativeBase] | MetaData): Any SQLAlchemy base, registry, or MetaData object. Must expose a
                `.metadata` attribute or be a MetaData instance.
            table_name (str): Name of the deduplication index table.
        """
        if isinstance(base_or_metadata, MetaData):
            metadata = base_or_metadata
        elif hasattr(base_or_metadata, "metadata"):
            metadata = base_or_metadata.metadata
        else:
            raise TypeError("Expected SQLAlchemy DeclarativeBase, registry, or MetaData object")

        if engine is None and session_factory is None:
            raise ValueError("Must provide either engine or session_factory")

        if session_factory is None:
            session_factory = sessionmaker(engine)

        self._session_factory = session_factory
        self._metadata = metadata
        self._table = Table(
            table_name,
            self._metadata,
            Column("band_idx", SmallInteger, nullable=False),
            Column("band_hash", BIGINT, nullable=False),
            Column("cluster_uuid", Uuid, nullable=False),
        )

        self._table.append_constraint(
            UniqueConstraint(
                "band_idx",
                "band_hash",
                name=f"{table_name}_band_idx_band_hash_key"
            )
        )

        # Precompile PostgreSQL insert stmt
        self._insert_sql = (
            pg_insert(self._table)
            .on_conflict_do_nothing(index_elements=["band_idx", "band_hash"])
        )

    def insert(self, bands: Iterable[tuple[int, int]]) -> UUID:
        check_existing_stmt = (
            select(self._table.c.cluster_uuid)
            .where(tuple_(self._table.c.band_idx, self._table.c.band_hash).in_(bands))
            .limit(1)
        )

        with self._session_factory() as session:
            # Commit returns before WAL is flushed to durable storage.
            # The transaction is visible immediately, but a crash before a WAL flush may
            # forget some committed transactions.
            #
            # This is a good-tradeoff as it saves nearly 5 ms of commit time for a high-volume
            # transaction in exchange for long-tail errors.
            _ = session.execute(text("SET LOCAL synchronous_commit = OFF"))

            existing_uuid = session.execute(check_existing_stmt).scalars().first()
            cluster_uuid = existing_uuid or uuid4()

            # NOTE: Only works on Postgres.
            values = [
                {"band_idx": i, "band_hash": h, "cluster_uuid": str(cluster_uuid)}
                for i, h in bands
            ]

            _ = session.execute(self._insert_sql, values)
            session.commit()

        return cluster_uuid

    def query(self, index: int, band: int) -> UUID | None:
        stmt = select(self._table.c.cluster_uuid).where(
            self._table.c.band_idx == index,
            self._table.c.band_hash == band,
        ).limit(1)

        with self._session_factory() as session:
            result = session.execute(stmt).scalars().first()

        return result



from abc import ABC, abstractmethod
from collections.abc import Iterable
from uuid import UUID, uuid4


class Backend(ABC):
    @abstractmethod
    def insert(self, bands: Iterable[tuple[int, int]]) -> UUID:
        pass

    @abstractmethod
    def query(self, index: int, band: int) -> UUID | None:
        pass


class LocalBackend(Backend):
    def __init__(self) -> None:
        """
        A local backend as an example of how to implement the Backend class.
        """
        self._index: dict[tuple[int, int], UUID] = {}

    def insert(self, bands: Iterable[tuple[int, int]]) -> UUID:
        found_uuid = None

        for index, band in bands:
            if (query := self.query(index, band)) is not None:
                found_uuid = query
                break

        if found_uuid is None:
            found_uuid = uuid4()
            for item in bands:
                self._index[item] = found_uuid

        return found_uuid

    def query(self, index: int, band: int) -> UUID | None:
        item = (index, band)

        if item in self._index:
            return self._index[item]

        return None



def n_grams(text: str, n: int = 3) -> list[str]:
    """
    Return a list of n‑grams for the supplied string.
    
    Args:
        text (str): The text string to find n-grams for
        n (int): The parameter which determines the length of the n-grams
    """
    if n <= 0:
        raise ValueError("n must be a positive integer")

    if len(text) < n:
        return []

    return [text[i : i + n] for i in range(len(text) - n + 1)]




import hashlib
import struct
import numpy as np
import xxhash
from collections.abc import Iterable
from uuid import UUID

from .backend import Backend, LocalBackend

MAX64 = np.uint64((1 << 64) - 1)
MIX_CONST = np.uint64(0x9e3779b97f4a7c15)


class DedupIndex:
    def __init__(
        self,
        backend: Backend | None = None,
        num_perms: int = 128,
        rows: int = 4,
    ) -> None:
        """
        Indexing layer that allows for query-time deduplication through hashing.

        Args:
            num_perms (int): The number of permutation functions to use to generate item signatures.
            rows (int): The number of rows to use when making signature bands.
        """
        self.num_hashes = num_perms
        self.rows = rows
        self.num_bands = num_perms // rows

        self._backend = LocalBackend() if backend is None else backend

    def _token_hash(self, token: str, seeds: np.ndarray) -> np.ndarray:
        """
        Hash computation for all seeds per token.

        Args:
            token (str): The token to hash.
            seeds (list[int]): The array of integer seeds to use for each hash row.

        Returns:
            list[int]: An array of integer has hash values, one per seed.
        """
        base = token.encode()
        hashes = np.empty(len(seeds), dtype=np.uint64)

        for i, seed in enumerate(seeds):
            # Concatenate without formatting overhead
            payload = base + b"-" + str(int(seed)).encode()
            hashes[i] = xxhash.xxh3_64(payload).intdigest()

        return hashes

    def _minhash_signature(self, tokens: Iterable[str]) -> np.ndarray:
        """
        Optimized MinHash calculation using numpy vectorization.

        Args:
            tokens (Iterable[str]): A list of tokens to compute the MinHash signature of.

        Returns:
            list[int]: A MinHash signature consisting of `num_rows` hashes.
        """
        tokens = list(tokens)
        seeds = np.arange(self.num_hashes, dtype=np.uint64)
        perm = seeds * MIX_CONST
        min_hashes = np.full(self.num_hashes, MAX64, dtype=np.uint64)

        for token in tokens:
            h0 = np.uint64(xxhash.xxh3_64(token).intdigest())
            token_hashes = h0 ^ perm
            min_hashes = np.minimum(min_hashes, token_hashes)

        return min_hashes

    def bands(self, tokens: Iterable[str]) -> list[int]:
        """
        Returns LSH bands. Currently, this only supports MinHash, which is the most popular algorithm for deduplicating
        large-scale data.

        Args:
            tokens (Iterable[str]): An iterable of any byte-encodeable objects which represent the content to
                deduplicate.

        Returns:
            list[str]: LSH bands derived from the MinHash signature of the tokens.
        """
        signature = self._minhash_signature(tokens)
        band_hashes: list[int] = []

        for i in range(0, len(signature), self.rows):
            band = signature[i:i + self.rows]
            payload = struct.pack(f"{len(band)}Q", *band)
            band_hash = np.uint64(xxhash.xxh64(payload).intdigest()).view(np.int64)
            band_hashes.append(int(band_hash))

        return band_hashes

    def items(self, bands: Iterable[int]) -> list[tuple[int, int]]:
        """
        A helper function which converts a list of bands to a normalized (index, band) format. Useful for inserting rows
        into a database.

        Args:
            bands (Iterable[str]): An iterable of LSH bands

        Returns:
            list[tuple[int, str]]: A list of tuples containing the index of the band hash, and then the band hash
                itself. Note that the band hash can be represented as a BIGINT
        """
        return [(idx, bh) for idx, bh in enumerate(bands)]

    def index(self, items: Iterable[tuple[int, int]]) -> UUID:
        """
        Retrieves the cluster UUID4 of a given items list derived from MinHash bands. This may add a new entry to the
        backend if the bands do not exist.

        Args:
            items (Iterable[str]): A list of item tuples. See `DedupIndex.items` for details.

        Returns:
            UUID: The cluster ID of the given MinHash bands.
        """
        return self._backend.insert(items)

    def query(self, tokens: Iterable[str]) -> UUID:
        """
        Retrieves the cluster UUID4 of the given tokens. This may add a new entry to the backend if the bands do not
        exist.

        Args:
            tokens (Iterable[str]): A list of tokens derived from some function such as the n_grams function.

        Returns:
            UUID: The cluster ID of the given tokens.
        """
        bands = self.bands(tokens)
        items = self.items(bands)
        return self.index(items)
